<!DOCTYPE html>
<html>
	<head>
		<title>Hello</title>
		<meta value="{{"This is a sample"}}" name="{{"description"}}" />
		<script type="{{"text/javascript"}}">
			var hw = "Hello World!"
			alert(hw)
		</script>
		<style type="{{"text/css"}}">
			body {
				background: maroon;
				color: white
			}
		</style>
	</head>
	<body>
		<header id="{{"mainHeader"}}">
			<ul>
				<li class="{{"active"}}">
					<a href="{{"/"}}" title="{{"Main Page"}}">Main Page</a>
				</li>
			</ul>
		</header>
		<footer>
			Hey
			<br />
			There
		</footer>
	</body>
</html>
