<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html>
	<head>
		<title>Hello</title>
		<meta name="{{"description"}}" value="{{"This is a sample"}}" />
		<meta content="{{"These are added by the child template"}}" name="{{"keywords"}}" />
	</head>
	<body>
		<header id="{{"mainHeader"}}">
			<ul>
				<li>Item 1</li>
				<li>Item 2</li>
			</ul>
		</header>
		<div id="{{"content"}}">
			<p>Content from child template</p>
		</div>
	</body>
</html>
